const YoutubeDL = require('youtube-dl');
const ytdl = require('ytdl-core');


//My Music module ;)
/**
 * Takes a discord.js client and turns it into a music bot.
 *
 * @param {Client} client - The discord.js client.
 * @param {object} options - (Optional) Options to configure the music bot. Acceptable options are:
 * 							prefix: The prefix to use for the commands (default '!').
 * 							global: Whether to use a global queue instead of a server-specific queue (default false).
 * 							maxQueueSize: The maximum queue size (default 20).
 * 							anyoneCanSkip: Allow anybody to skip the song.
 * 							clearInvoker: Clear the command message.
 * 							volume: The default volume of the player.
 */
module.exports = function (client, options) {
	// Get all options.
	let PREFIX = (options && options.prefix) || '!';
	let GLOBAL = (options && options.global) || false;
	let MAX_QUEUE_SIZE = (options && options.maxQueueSize) || 100;
	let DEFAULT_VOLUME = (options && options.volume) || 50;
	let ALLOW_ALL_SKIP = (options && options.anyoneCanSkip) || false;
	let CLEAR_INVOKER = (options && options.clearInvoker) || false;

	// Create an object of queues.
	let queues = {};
	var somestuff = {};

	// Catch message events.
	client.on('message', msg => {
		const message = msg.content.trim();

		// Check if the message is a command.
		if (message.toLowerCase().startsWith(PREFIX.toLowerCase())) {
			// Get the command and suffix.
			const command = message.substring(PREFIX.length).split(/[ \n]/)[0].toLowerCase().trim();
			const suffix = message.substring(PREFIX.length + command.length).trim();
			if (!somestuff[message.guild.id]) somestuff[message.guild.id] = {
			  loop: 0,
			}
			// Process the commands.
			switch (command) {
				case 'play':
				if(message.channel.type === 'dm') return msg.channel.send('\`' + command + ' \` can only be used in a guild/server')
					return play(msg, suffix);
				case 'p':
				if(message.channel.type === 'dm') return msg.channel.send('\`' + command + ' \` can only be used in a guild/server')
					return play(msg, suffix);
				case 'search':
				if(message.channel.type === 'dm') return msg.channel.send('\`' + command + ' \` can only be used in a guild/server')
					return play(msg, suffix);
				case 'skip':
				if(message.channel.type === 'dm') return msg.channel.send('\`' + command + ' \` can only be used in a guild/server')
					return skip(msg, suffix);
				case 'queue':
				if(message.channel.type === 'dm') return msg.channel.send('\`' + command + ' \` can only be used in a guild/server')
					return queue(msg, suffix);
				case 'pause':
				if(message.channel.type === 'dm') return msg.channel.send('\`' + command + ' \` can only be used in a guild/server')
					return pause(msg, suffix);
				case 'resume':
				if(message.channel.type === 'dm') return msg.channel.send('\`' + command + ' \` can only be used in a guild/server')
					return resume(msg, suffix);
				case 'volume':
				if(message.channel.type === 'dm') return msg.channel.send('\`' + command + ' \` can only be used in a guild/server')
					return volume(msg, suffix);
				case 'leave':
				if(message.channel.type === 'dm') return msg.channel.send('\`' + command + ' \` can only be used in a guild/server')
					return leave(msg, suffix);
				case 'stop':
					return clearqueue(msg, suffix);
				case 'clearqueue':
				if(message.channel.type === 'dm') return msg.channel.send('\`' + command + ' \` can only be used in a guild/server')
					return clearqueue(msg, suffix);
				case "loop":
					return loop(msg, suffix);
			}
			if (CLEAR_INVOKER) {
				msg.delete();
			}
		}
	});

	/**
	 * Checks if a user is an admin.
	 *
	 * @param {GuildMember} member - The guild member
	 * @returns {boolean} -
	 */
	function isAdmin(member) {
		return member.hasPermission("ADMINISTRATOR");
	}

	/**
	 * Checks if the user can skip the song.
	 *
	 * @param {GuildMember} member - The guild member
	 * @param {array} queue - The current queue
	 * @returns {boolean} - If the user can skip
	 */
	function canSkip(member, queue) {
		if (ALLOW_ALL_SKIP) return true;
		else if (queue[0].requester === member.id) return true;
		else if (isAdmin(member)) return true;
		else return false;
	}

	/**
	 * Gets the song queue of the server.
	 *
	 * @param {integer} server - The server id.
	 * @returns {object} - The song queue.
	 */
	function getQueue(server) {
		// Check if global queues are enabled.
		if (GLOBAL) server = '_'; // Change to global queue.

		// Return the queue.
		if (!queues[server]) queues[server] = [];
		return queues[server];
	}

	/**
	 * The command for adding a song to the queue.
	 *
	 * @param {Message} msg - Original message.
	 * @param {string} suffix - Command suffix.
	 * @returns {<promise>} - The response edit.
	 */
	function play(msg, suffix) {
		// Make sure the user is in a voice channel.
		if (msg.member.voiceChannel === undefined) return msg.channel.sendMessage(':x: You\'re not in a voice channel.');

		// Make sure the suffix exists.
		if (!suffix) return msg.channel.sendMessage(':x: Please provide a link/title of a song for me to play');

		// Get the queue.
		const queue = getQueue(msg.guild.id);

		// Check if the queue has reached its maximum size.
		if (queue.length >= MAX_QUEUE_SIZE) {
			return msg.channel.sendMessage('Maximum queue size reached!');
		}

		// Get the video information.
		msg.channel.sendMessage(':mag_right: **Searching**  `' + suffix + '` . . .').then(response => {
			var searchstring = suffix
			if (!suffix.toLowerCase().startsWith('http')) {
				searchstring = 'gvsearch1:' + suffix;
			}

			YoutubeDL.getInfo(searchstring, ['-q', '--no-warnings', '--force-ipv4'], (err, info) => {
				// Verify the info.
				if (err || info.format_id === undefined || info.format_id.startsWith('0')) {
					return response.edit('```Sorry I could not find that certian video!```');
				}

				info.requester = msg.author.id;

				// Queue the video.
				response.edit(':musical_note: Added **' + info.title + '** to the queue!').then(() => {
					queue.push(info);
					// Play if only one element in the queue.
					if (queue.length === 1) executeQueue(msg, queue);
				}).catch(console.log);
			});
		}).catch(console.log);
	}


	function skip(msg, suffix) {
		// Get the voice connection.
		const voiceConnection = client.voiceConnections.find(val => val.channel.guild.id == msg.guild.id);
		if (voiceConnection === null) return msg.channel.sendMessage(':x: No music being played.');

		// Get the queue.
		const queue = getQueue(msg.guild.id);

		if (!canSkip(msg.member, queue)) return msg.channel.sendMessage(':x: You cannot skip this as you didn\'t queue it.').then((response) => {
			response.delete(5000);
		});

		// Get the number to skip.
		let toSkip = 1; // Default 1.
		if (!isNaN(suffix) && parseInt(suffix) > 0) {
			toSkip = parseInt(suffix);
		}
		toSkip = Math.min(toSkip, queue.length);

		// Skip.
		queue.splice(0, toSkip - 1);

		// Resume and stop playing.
		const dispatcher = voiceConnection.player.dispatcher;
		if (voiceConnection.paused) dispatcher.resume();
		dispatcher.end();

		msg.channel.sendMessage('**Skipped** ' + toSkip + '!');
	}

	function queue(msg, suffix) {
		// Get the queue.
		const queue = getQueue(msg.guild.id);

		// Get the queue text.
		const text = queue.map((video, index) => (
			(index + 1) + ': ' + video.title
		)).join('\n');

		// Get the status of the queue.
		let queueStatus = 'Stopped';
		const voiceConnection = client.voiceConnections.find(val => val.channel.guild.id == msg.guild.id);
		if (voiceConnection !== null) {
			const dispatcher = voiceConnection.player.dispatcher;
			queueStatus = dispatcher.paused ? 'Paused' : 'Playing';
		}

		// Send the queue and status.
		msg.channel.sendMessage('```Queue``` \(' + queueStatus + '\):\n' + '`' + text + '`');
	}

	/**
	 * The command for pausing the current song.
	 *
	 * @param {Message} msg - Original message.
	 * @param {string} suffix - Command suffix.
	 * @returns {<promise>} - The response message.
	 */
	function pause(msg, suffix) {
		// Get the voice connection.
		const voiceConnection = client.voiceConnections.find(val => val.channel.guild.id == msg.guild.id);
		if (voiceConnection === null) return msg.channel.sendMessage(':x: No music being played.');

		if (!isAdmin(msg.member))
			return msg.channel.sendMessage(':x: Sorry you do not have the permission to use this command');

		// Pause.
		msg.channel.sendMessage('Playback paused.');
		const dispatcher = voiceConnection.player.dispatcher;
		if (!dispatcher.paused) dispatcher.pause();
	}

	/**
	 * The command for leaving the channel and clearing the queue.
	 *
	 * @param {Message} msg - Original message.
	 * @param {string} suffix - Command suffix.
	 * @returns {<promise>} - The response message.
	 */
	function leave(msg, suffix) {
		if (isAdmin(msg.member)) {
			const voiceConnection = client.voiceConnections.find(val => val.channel.guild.id == msg.guild.id);
			if (voiceConnection === null) return msg.channel.sendMessage(':x: I\'m not in any channel!.');
			// Clear the queue.
			somestuff[message.guild.id].loop=0;
			const queue = getQueue(msg.guild.id);
			queue.splice(0, queue.length);

			// End the stream and disconnect.
			voiceConnection.player.dispatcher.end();
			voiceConnection.disconnect();
		} else {
			msg.channel.sendMessage(':x: You don\'t have permission to use this command!');
		}
	}

	function clearqueue(msg, suffix) {
		if (isAdmin(msg.member)) {
			somestuff[message.guild.id].loop=0;
			const queue = getQueue(msg.guild.id);

			queue.splice(0, queue.length);
			msg.channel.sendMessage('Queue cleared!');
		} else {
			msg.channel.sendMessage(':x: You don\'t have permission to use that command!');
		}
	}

	/**
	 * The command for resuming the current song.
	 *
	 * @param {Message} msg - Original message.
	 * @param {string} suffix - Command suffix.
	 * @returns {<promise>} - The response message.
	 */
	function resume(msg, suffix) {
		// Get the voice connection.
		const voiceConnection = client.voiceConnections.find(val => val.channel.guild.id == msg.guild.id);
		if (voiceConnection === null) return msg.channel.sendMessage(':x: No music being played.');

		if (!isAdmin(msg.member))
			return msg.channel.sendMessage(':x: You don\'t have permission to use that command!');

		// Resume.
		msg.channel.sendMessage('Playback resumed.');
		const dispatcher = voiceConnection.player.dispatcher;
		if (dispatcher.paused) dispatcher.resume();
	}

	function volume(msg, suffix) {
		// Get the voice connection.
		const voiceConnection = client.voiceConnections.find(val => val.channel.guild.id == msg.guild.id);
		if (voiceConnection === null) return msg.channel.sendMessage(':x: No music being played.');

		if (!isAdmin(msg.member))
			return msg.channel.sendMessage(':x: You don\'t have permission to use that command!');

		// Get the dispatcher
		const dispatcher = voiceConnection.player.dispatcher;

		if (suffix > 200 || suffix < 0) return msg.channel.sendMessage('The number you entered was too big, max is 200').then((response) => {
			response.delete(5000);
		});

		msg.channel.sendMessage("Volume set to " + suffix);
		dispatcher.setVolume((suffix/100));
	}

	function loop(msg, suffix) {
		if (suffix === "on") {
			msg.channel.send(":white_check_mark: loop turned on!")
			somestuff[message.guild.id].loop=1;
		} else if (suffix === "off") {
			msg.channel.send(":white_check_mark: loop turned off!")
			somestuff[message.guild.id].loop=0;
		} else {
			msg.channel.send(":x: usage: loop on/off")
		}
	}
	function executeQueue(msg, queue) {
		// If the queue is empty, finish.
		if (queue.length === 0) {
			msg.channel.sendMessage('```Queue Ended```');

			// Leave the voice channel.
			const voiceConnection = client.voiceConnections.find(val => val.channel.guild.id == msg.guild.id);
			if (voiceConnection !== null) return voiceConnection.disconnect();
		}

		new Promise((resolve, reject) => {
			// Join the voice channel if not already in one.
			const voiceConnection = client.voiceConnections.find(val => val.channel.guild.id == msg.guild.id);
			if (voiceConnection === null) {
				// Check if the user is in a voice channel.
				if (msg.member.voiceChannel) {
					msg.member.voiceChannel.join().then(connection => {
						resolve(connection);
					}).catch((error) => {
						console.log(error);
					});
				} else {
					// Otherwise, clear the queue and do nothing.
					queue.splice(0, queue.length);
					reject();
				}
			} else {
				resolve(voiceConnection);
			}
		}).then(connection => {
			// Get the first item in the queue.
			const video = queue[0];

			// Play the video.
			msg.channel.sendMessage(':notes: **Now Playing** **' + video.title + '** :notes:').then(() => {
				let dispatcher = connection.playStream(ytdl(video.webpage_url, {filter: 'audioonly'}), {seek: 0, volume: (DEFAULT_VOLUME/100)});

				connection.on('error', (error) => {
					// Skip to the next song.
					console.log(error);
					queue.shift();
					if (somestuff[message.guild.id].loop === 1) {queue.push(video.webpage_url)}
					executeQueue(msg, queue);
				});

				dispatcher.on('error', (error) => {
					// Skip to the next song.
					console.log(error);
					queue.shift();
					if (somestuff[message.guild.id].loop === 1) {queue.push(video.webpage_url)}
					executeQueue(msg, queue);
				});

				dispatcher.on('end', () => {
					// Wait a second.
					setTimeout(() => {
						if (queue.length > 0) {
							// Remove the song from the queue.
							queue.shift();
							if (somestuff[message.guild.id].loop === 1) {queue.push(video.webpage_url)}
							// Play the next song in the queue.
							executeQueue(msg, queue);
						}
					}, 1000);
				});
			}).catch((error) => {
				console.log(error);
			});
		}).catch((error) => {
			console.log(error);
		});
	}
}
